unit hiMathParseEx;

interface

{$I share.inc}

uses Kol,
     {$ifndef F_P}err,{$endif}
     Share,Debug;

type
  TProgram = class
   private
    FPgm: Pointer;
    FAlloced: integer;
    FUsed: integer;
    FPos: integer;
    
    FEOS: boolean; 
    
    procedure SetPosition(val: integer);
    
    function readBinary(var v; cnt:integer): boolean;
    function writeBinary(var v; cnt:integer): boolean;
    
    function AllocMemIfNeeded(need:integer): boolean;
   public
    constructor Create;
    destructor Destroy; override;
    
    procedure Clear;
    procedure Truncate; //  
    
    function readByte(var v: byte): boolean;
    function readInteger(var v: integer): boolean;
    function readReal(var v: real): boolean;
    function readString(var v: string): boolean;
    
    function writeByte(v: byte): boolean;
    function writeInteger(v: integer): boolean;
    function writeReal(v: real): boolean;
    function writeString(v: string): boolean;
    
    property EOS:boolean read FEOS; //end of stream
    property Position:integer read FPos write SetPosition;
    property Length:integer read FUsed;
  end;
  
  TStack = class
    private
      FArr: array of real;
      FSize: integer;
      FCurrent: integer;
      
      procedure AllocMoreMem;
    public
      constructor Create;
      destructor Destroy; override;
      
      procedure SetSize(cnt: integer);
      procedure Push(val: real);
      function Pop(var val: real): boolean;
      procedure Clear;
      
      property Size: integer read FSize;
      property CurrentIndex: integer read FCurrent;
  end;

  THIMathParseEx = class(TDebug)
   private
    FProgram: TProgram;
    FStack: TStack;
    FDebug: boolean;
   
    FDataCount:integer;
    Token:string;
    TokType:byte;
    Line:string;
    SafeLine:string;
    LPos:smallint;
    FResult:real;
    FDefault:real;
    AngleMode:real;
    Err:smallint;
    dt:TData;
    Args:array of real;
    flags:array of boolean;

    procedure MakeExpressionProgram;
    procedure ExecuteExpression;
    procedure ExecuteExpressionPart;
    procedure CalcFunc;

    procedure SetCount(Value:integer);
    procedure SetLine(Value:string);
    procedure SetDefault(Value:real);
    procedure SetAngleMode(Value:byte);
    procedure SetDebugMode(value:boolean);
    function CalcErrPos(Value:integer): string;

    procedure GetToken;

    procedure Level0;
    procedure Level1a; // < > <= >= =
    procedure Level1b; // + -
    procedure Level2; // / *
    procedure Level3; // ^
    procedure Level4; // unar + - not
    procedure Level5; // and or xor
    procedure Level6; // ( function
    function ReadFunc(f:string):boolean;
   public
    X:array of THI_Event;
    _event_onError:THI_Event;
    _event_onResult:THI_Event;
    _prop_ResultType:byte;
    _prop_ExtNames:integer;

    constructor Create;
    destructor Destroy; override;
    procedure _work_doCalc(var _Data:TData; Index:word);
    procedure _work_doMathStr(var _Data:TData; Index:word);
    procedure _work_doClear(var _Data:TData; Index:word);
    procedure _work_doAngleMode(var _Data:TData; Index:word);
    procedure _work_doDefault(var _Data:TData; Index:word);
    procedure _var_Result(var _Data:TData; Index:word);
    procedure _var_reCalc(var _Data:TData; Index:word);
    procedure _var_PosErr(var _Data:TData; Index:word);
    property _prop_DataCount:integer write SetCount;
    property _prop_MathStr:string write SetLine;
    property _prop_Default:real write SetDefault;
    property _prop_AngleMode:byte write SetAngleMode;
    property _prop_DebugMode:boolean read FDebug write SetDebugMode;
  end;

procedure Power(var x, Exponent: real);
function IntPower(const Base:Extended; const Exponent:Integer):Extended;
function ArcTan2(const Y,X:Extended):Extended;
function LogN(const Base,X:Extended):Extended;
function Tan(const X: Extended): Extended;

implementation

//------------------------------ TProgram -------------------------------------- 

function ShiftPointer(p: Pointer; cnt: integer): Pointer;
begin
  Result := Pointer(Integer(p) + cnt);
end;

constructor TProgram.Create;
begin
  inherited Create;
  FPgm := nil;
  FAlloced := 0;
  FUsed := 0;
  FPos := 0;
  FEOS := True;
end;

destructor TProgram.Destroy;
begin
  if FPgm <> nil then begin
    FreeMem(FPgm, FAlloced);
  end;
  inherited Destroy;
end;

procedure TProgram.SetPosition(val: integer);
begin
  if val > FUsed then begin
    FPos := FUsed - 1;
  end else if val < 0 then begin
    FPos := 0;
  end else begin
    FPos := val;
  end;
end;

function TProgram.readBinary(var v; cnt:integer): boolean;
begin
  Result := False;
  if FPos + cnt > FUsed then begin
    Exit;
  end;
  Move(ShiftPointer(FPgm, FPos)^, v, cnt);
  FPos := FPos + cnt;
  Result := True;
end;

function TProgram.writeBinary(var v; cnt:integer): boolean;
begin
  Result := False;
  if not AllocMemIfNeeded(cnt) then begin
    Exit;
  end;
  Move(v, ShiftPointer(FPgm, FPos)^, cnt);
  FPos := FPos + cnt;
  Result := True;
end;

function TProgram.AllocMemIfNeeded(need:integer): boolean;
var nsz: integer;
begin
  nsz := FAlloced;
  while FPos + need > nsz do begin
    nsz := nsz + 100;
  end;
  if nsz <> FAlloced then begin
    ReallocMem(FPgm, nsz);
  end;
  if FUsed < FPos + need then begin
    FUsed := FPos + need;
  end;
  Result := FPgm <> nil;
end;

procedure TProgram.Clear;
begin
  FUsed := 0;
  FPos := 0;
  FreeMem(FPgm, FAlloced);
  FPgm := nil;
  FAlloced := 0;
end;

procedure TProgram.Truncate;
begin
  if FUsed < FAlloced then begin
    ReallocMem(FPgm, FUsed);
    FAlloced := FUsed;
  end;
end;

function TProgram.readByte(var v: byte): boolean;
begin
  Result := ReadBinary(v, sizeof(Byte));
end;

function TProgram.readInteger(var v: integer): boolean;
begin
  Result := ReadBinary(v, sizeof(Integer));
end;

function TProgram.readReal(var v: real): boolean;
begin
  Result := ReadBinary(v, sizeof(Real));
end;

function TProgram.readString(var v: string): boolean;
var l: integer;
begin
  if ReadInteger(l) then begin
    if l < 0 then l := 0;
    SetLength(v, l);
    Result := (l = 0) or (ReadBinary(v[1], l));
  end else
    Result := False;
end;

function TProgram.writeByte(v: byte): boolean;
begin
  Result := WriteBinary(v, sizeof(Byte));
end;

function TProgram.writeInteger(v: integer): boolean;
begin
  Result := WriteBinary(v, sizeof(Integer));
end;

function TProgram.writeReal(v: real): boolean;
begin
  Result := WriteBinary(v, sizeof(Real));
end;

function TProgram.writeString(v: string): boolean;
var l: integer;
begin
  l := System.Length(v);
  if WriteInteger(l) then begin
    if WriteBinary(v[1], l) then begin
      Result := True;
    end else begin
      FPos := FPos - sizeof(Integer);
      Result := False;
    end;
  end else
    Result := False;
end;

//------------------------------ TProgram --------------------------------------
//------------------------------- TStack ---------------------------------------

constructor TStack.Create;
begin
  inherited Create;
  FCurrent := 0;
  FSize := 0;
end;

destructor TStack.Destroy;
begin
  if FSize > 0 then begin
    SetLength(FArr, 0);
  end;
  inherited Destroy;
end;

procedure TStack.AllocMoreMem;
begin
  FSize := FSize + 20;
  SetLength(FArr, FSize);
end;

procedure TStack.SetSize(cnt: integer);
begin
  FSize := cnt;
  if FSize < 0 then FSize := 0;
  SetLength(FArr, FSize);
  if FCurrent > FSize then
    FCurrent := FSize;
end;

procedure TStack.Push(val: real);
begin
  if FCurrent >= FSize then begin
    AllocMoreMem;
  end;
  FArr[FCurrent] := val;
  inc(FCurrent);
end;

function TStack.Pop(var val: real): boolean;
begin
  if FCurrent <= 0 then begin
    Result := False;
    Exit;
  end;
  dec(FCurrent);
  val := FArr[FCurrent];
  Result := True;
end;

procedure TStack.Clear;
begin
  FCurrent := 0;
end;
      
//------------------------------- TStack ---------------------------------------

const
  //  []     
  //  (+,-,*,...)    
  CmdNop     = $00;
  CmdPush    = $01; // [val:real]  
  CmdRead    = $02; // [n:byte]  %n
  CmdEnd     = $03; //  
  CmdDbgLine = $04; // [pos:integer]    
  
  CmdAdd     = $10; // +
  CmdSub     = $11; // -
  CmdMul     = $12; // *
  CmdDiv     = $13; // /
  CmdBig     = $14; // >
  CmdSmall   = $15; // <
  CmdEBig    = $16; // >=
  CmdESmall  = $17; // <=
  CmdEqual   = $18; // ==
  CmdNEqual  = $19; // <>
  CmdNeg     = $1A; // unar -
  CmdPow     = $1B; // ^
  CmdNot     = $1C; // not
  CmdAnd     = $1D; // and
  CmdOr      = $1E; // or
  CmdXor     = $1F; // xor
  CmdIDiv    = $20; // div
  CmdIMod    = $21; // mod
  
  CmdFunc    = $50; // [n:byte] math func n
  CmdArr     = $51; // [n:byte]    %n
  CmdMatr    = $52; // [n:byte]    %n
  CmdUFunc   = $53; // [n:byte] %n()
  CmdUNFunc  = $54; // [name:string] name()

const
  TokName   = 1;
  TokReal   = 2;
  TokNumber = 3;
  TokSymbol = 4;
  TokMath   = 5;
  TokEnd    = 6;
  TokHex    = 7;
  digE      = 2.718281828459045; //may be defined in Delphi
  serr      = 'MathParse Exception';
  s:array[0..1]of string = ('    MathParserEx'#13#10'   '
                           ,'    MathParserEx'#13#10'   ');

constructor THIMathParseEx.Create;
begin
  inherited Create;
  FProgram := TProgram.Create;
  FStack := TStack.Create;
end;

destructor THIMathParseEx.Destroy;
begin
  FProgram.Free;
  FStack.Free;
  inherited Destroy;
end;

procedure THIMathParseEx.MakeExpressionProgram;
begin
  FProgram.Clear;
  {$ifndef F_P}
  try
  {$endif}
    Level0;
  {$ifndef F_P}
  except on E:Exception do
    case E.Code of
      e_Custom: Err:=0;//SyntaxError
      else      Err:=1;//CalcError
    end;//case
  end;//except
  {$endif}
  if Err < 0 then begin
    LPos:=-1;
    FProgram.Truncate;
  end else begin
    FProgram.Clear;
    dec(LPos);
    if assigned(_event_onError.Event) then
      _hi_onEvent(_event_onError, Err)
    else _debug(s[Err]+ CalcErrPos(LPos));
  end;
end;

procedure THIMathParseEx.ExecuteExpression;
var i: integer;
begin
  for i:=0 to FDataCount-1 do Flags[i] := false;
  if LPos>=0 then FResult:=FDefault;
  LPos := 1; Err := -1;
  FProgram.Position := 0;
  FStack.Clear;
  ExecuteExpressionPart;
  if (Err < 0) then FStack.Pop(FResult);
end;

procedure THIMathParseEx.ExecuteExpressionPart;
var cmd: byte;
    r, x, x1, x2, y: real;
    b: boolean;
    Fd, FItem: TData;
    tmp: PData;
    i: byte;
    cnt, mx, my: integer;
    s: string;
    Arr: PArray;
    Mtx: PMatrix;
begin
  while FProgram.ReadByte(cmd) do begin
    case cmd of
      CmdNop: ;
      CmdEnd: Exit; //  
      
      CmdPush: // [val:real]  
      begin
        FProgram.ReadReal(r);
        FStack.Push(r);
      end;
      
      CmdRead: // [n:byte]  %n
      begin
        FProgram.ReadByte(i);
        if i = 255 then r := FResult
        else if Flags[i] then r := Args[i] //   
        else begin
          Fd := ReadData(dt,self.X[i],nil);
          if _IsNULL(Fd) and (Fd.sdata=serr) then
            begin Err := Fd.idata; exit; end;
          r := ToReal(Fd);
          //  ,      
          Args[i] := r;
          Flags[i] := true;
        end;
        FStack.Push(r);
      end;
      
      CmdDbgLine: // [pos:integer]    
      begin
        FProgram.ReadInteger(cnt);
        LPos := cnt;
      end;

      // + - * / ^
      CmdAdd, CmdSub, CmdMul, CmdDiv, CmdPow, CmdIDiv, CmdIMod:
      begin
        FStack.Pop(x2);
        FStack.Pop(x1);
        case cmd of
          CmdAdd: r := x1 + x2;
          CmdSub: r := x1 - x2;
          CmdMul: r := x1 * x2;
          CmdDiv: if x2 = 0 then begin Err := 1; Exit; end else r := x1 / x2;
          CmdIDiv: if Round(x2) = 0 then begin Err := 1; Exit; end else r := Round(x1) div Round(x2);
          CmdIMod: if Round(x2) = 0 then begin Err := 1; Exit; end else r := Round(x1) mod Round(x2);
          CmdPow: if (Round(x2) <> x2) and (x1 <= 0) then begin Err:=1; Exit; end else begin
            Power(x1, x2);
            r := x1;
          end;
        end;
        FStack.Push(r);
      end;
      
      // > < >= <= == <> and or xor
      CmdBig, CmdSmall, CmdEBig, CmdESmall, CmdEqual, CmdNEqual,
      CmdAnd, CmdOr, CmdXor:
      begin
        FStack.Pop(x2);
        FStack.Pop(x1);
        case cmd of
          CmdBig: b := x1 > x2;
          CmdSmall: b := x1 < x2;
          CmdEBig: b := x1 >= x2;
          CmdESmall: b := x1 <= x2;
          CmdEqual: b := x1 = x2;
          CmdNequal: b := x1 <> x2;
          CmdAnd: b := (x1 <> 0) and (x2 <> 0);
          CmdOr: b := (x1 <> 0) or (x2 <> 0);
          CmdXor: b := (x1 <> 0) xor (x2 <> 0);
        end;
        FStack.Push(ord(b));
      end;
      
      CmdNeg, CmdNot: // unar - not
      begin
        FStack.Pop(x1);
        case cmd of
          CmdNeg: r := -x1;
          CmdNot: ;
        end;
        FStack.Push(r);
      end;
      
      CmdFunc: // [n:byte] math func n
      begin
        CalcFunc;
        if Err >= 0 then Exit;
      end;
      
      CmdArr: // [n:byte]    %n
      begin
        FProgram.ReadByte(i);
        Arr := ReadArray(self.X[i]);
        if Arr = nil then begin Err:=1; exit; end;
        FStack.Pop(y);
        dtReal(Fd, y);
        if not Arr._Get(Fd, FItem) then begin Err:=1; exit; end;
        FStack.Push(ToReal(FItem));
      end;
      
      CmdMatr: // [n:byte]    %n
      begin
        FProgram.ReadByte(i);
        Mtx := ReadMatrix(self.X[i]);
        if Mtx=nil then begin Err :=1 ; exit; end;
        FStack.Pop(y);
        FStack.Pop(x);
        mx := round(y); my := round(x);
        Fd := Mtx._Get(mx, my);
        if _IsNull(Fd) then begin Err:=1; exit; end;
        FStack.Push(ToReal(Fd));
      end;
      
      CmdUFunc: // [n:byte] %n()
      begin
        FProgram.ReadByte(i);
        FProgram.ReadInteger(cnt);
        dtNull(Fd);
        TRY
          if cnt > 0 then begin
            FStack.Pop(y);
            dtReal(Fd, y);
            dec(cnt);
            while cnt > 0 do begin  //   
              FStack.Pop(y);
              dtReal(FItem, y);
              AddMTData(@FItem, @Fd, tmp);
              FreeData(@Fd);
              Fd := FItem;
              dec(cnt);
            end;
          end;
          FItem := Fd;
          _ReadData(FItem,self.X[i]);
        FINALLY
          FreeData(@Fd);
        END;
        if _IsNULL(FItem) then begin Err := 1; exit; end;
        FStack.Push(ToReal(FItem));
      end;
      
      CmdUNFunc: // name()
      begin
        //     "  "
        FProgram.ReadString(s);
        FProgram.ReadInteger(cnt);
        dtNull(Fd);
        TRY
          if cnt > 0 then begin
            FStack.Pop(y);
            dtReal(Fd, y);
            dec(cnt);
            while cnt > 0 do begin  //   
              FStack.Pop(y);
              dtReal(FItem, y);
              AddMTData(@FItem, @Fd, tmp);
              FreeData(@Fd);
              Fd := FItem;
              dec(cnt);
            end;
          end;
          dtString(FItem, s);
          if not _IsNull(Fd) then begin
            AddMTData(@FItem, @Fd, tmp);
          end;
          _ReadData(FItem,self.X[_prop_ExtNames-1]);
        FINALLY
          FreeData(@Fd);
        END;
        if _IsNULL(FItem) then begin Err := 1; exit; end;
        FStack.Push(ToReal(FItem));
      end;
      
      else begin
        Err := 2;
        Exit;
      end;
    end;
  end;
end;

function THIMathParseEx.CalcErrPos;
var
  ListLine: PStrListEx;
  i: integer;
  s: string;
begin
  Result := '0:1';
  ListLine := NewStrListEx;
TRY
  s := SafeLine;
  Replace(s, #1, '');
  ListLine.SetText(s, false);
  if ListLine.Count = 0 then exit;
  i := 0;
  ListLine.Objects[0] := 1;
  if ListLine.Count > 1 then
    for i := 1 to ListLine.Count - 1 do
    begin
      ListLine.Objects[i] := Length(ListLine.Items[i - 1]) + integer(ListLine.Objects[i - 1]);
      if Value < integer(ListLine.Objects[i]) + Length(ListLine.Items[i]) then break;
    end;
  Result := int2str(Value - integer(ListLine.Objects[i]) + 1) + ':' + int2str(i + 1);
FINALLY
  ListLine.free;
END;
end;

procedure THIMathParseEx.SetLine;
begin
  Line := Value+#1;
  SafeLine := Line;
  Replace(Line, #13#10, '');
  MakeExpressionProgram;
end;

procedure THIMathParseEx.SetAngleMode;
begin
  if Value=0 then AngleMode :=1
  else AngleMode:=pi/180;
end;

procedure THIMathParseEx.SetDefault;
begin
  FDefault:=Value;
  FResult:=FDefault;
  LPos:=-1;
  Err :=-1;
end;

procedure THIMathParseEx.SetDebugMode;
begin
  FDebug := value;
  MakeExpressionProgram;
end;

procedure THIMathParseEx._work_doDefault;
begin
  SetDefault(ToReal(_Data));
end;

procedure THIMathParseEx._work_doAngleMode;
begin
  SetAngleMode(ToInteger(_Data));
end;

procedure THIMathParseEx._work_doClear;
begin
  FResult:=FDefault;
  LPos:=-1;
  Err :=-1;
end;

procedure THIMathParseEx.SetCount;
begin
  SetLength(X,Value);
  SetLength(Args,Value);
  SetLength(Flags,Value);
  FDataCount := Value;
end;

procedure THIMathParseEx._work_doCalc;
var Res:integer;
    x:real;
begin
  dt  := _Data;
  Res := 0;
  ExecuteExpression;
  if Err < 0 then begin
    LPos:=-1;
    if _prop_ResultType=0 then
      Res:=Round(FResult);
    if _prop_ResultType=0 then _hi_CreateEvent(_Data,@_event_onResult,Res)
    else _hi_CreateEvent(_Data,@_event_onResult,FResult)
  end else begin
    dec(LPos);
    if assigned(_event_onError.Event) then
      _hi_CreateEvent(_Data,@_event_onError,Err)
//    else _debug(s[Err]+ int2Str(LPos));
    else _debug(s[Err]+ CalcErrPos(LPos));
  end;
end;//_work_doCalc

procedure THIMathParseEx._var_reCalc;
var Y:real;
begin
  dt := _Data;
  ExecuteExpression;
  if Err < 0 then begin
    if _prop_ResultType=0 then dtInteger(_Data,Round(FResult))
    else dtReal(_Data,Y);
    LPos:=-1;
  end else begin
    dtNULL(_Data);
    _Data.idata := Err;
    _Data.sdata := serr;
   end; 
end;

procedure THIMathParseEx._work_doMathStr;
begin
  if _IsStr(_Data) then _prop_MathStr := ToString(_Data);
end;

procedure THIMathParseEx._var_Result;
begin
  if Err>=0 then begin
    dtNULL(_Data);
    _Data.idata := Err;
    _Data.sdata := serr;
   end
  else if _prop_ResultType = 0 then dtInteger(_Data,Round(FResult))
  else dtReal(_Data,FResult);
end;

procedure THIMathParseEx._var_PosErr;
begin
  dtInteger(_Data,LPos);
end;

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~ PARSE ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

procedure THIMathParseEx.GetToken;
begin
   Token := '';
   TokType := 0;
   while Line[LPos] in [' ',#9] do inc(LPos);
   if _prop_DebugMode then begin
       FProgram.WriteByte(CmdDbgLine);
       FProgram.WriteInteger(LPos);
   end;
   case Line[LPos] of
     'a'..'z','A'..'Z','_',''..'',''..'':
       begin
         repeat
           Token := Token + Line[LPos];
           inc(LPos);
         until not(Line[LPos] in ['a'..'z','A'..'Z','_',''..'',''..'','0'..'9']);
         //Token := LowerCase(Token);
         TokType := TokName;
       end;
     '.','$','0'..'9':
       begin
         if (Line[LPos]='$')or((Line[LPos]='0')and(Line[LPos+1]='x')) then
          begin
           if Line[LPos]<>'$' then inc(LPos);
           inc(LPos);
           while Line[LPos] in ['0'..'9','a'..'f','A'..'F'] do begin
             Token := Token + Line[LPos];
             inc(LPos)
            end;
           if Token<>'' then TokType := TokHex;
           exit;
          end;
         while Line[LPos] in ['0'..'9'] do begin
           Token := Token + Line[LPos];
           inc(LPos)
          end;
         TokType := TokNumber;
         if Line[LPos]='.' then
          begin
           Token := Token + '.';
           inc(LPos);
           while Line[LPos] in ['0'..'9'] do begin
             Token := Token + Line[LPos];
             inc(LPos);
            end;
           if Token='.' then begin
             TokType := 0; exit end;
           TokType := TokReal;
          end;
         while Line[LPos] in [' ',#9] do inc(LPos);
         if Line[LPos] in ['e','E'] then begin
           Token := Token + Line[LPos];
           inc(LPos);
           while Line[LPos] in [' ',#9] do inc(LPos);
           if Line[LPos] in ['+','-'] then begin
             Token := Token + Line[LPos];
             inc(LPos);
             while Line[LPos] in [' ',#9] do inc(LPos);
            end;
           if not(Line[LPos] in ['0'..'9']) then begin
             TokType := 0; exit end;
           while Line[LPos] in ['0'..'9'] do begin
             Token := Token + Line[LPos];
             inc(LPos);
            end;
           TokType := TokReal;
          end
       end;
     '(',')','%',',','[',']':
        begin Token := Line[LPos]; inc(LPos); TokType := TokSymbol end;
     '<','>','=':
        begin
          Token := Line[LPos];
          inc(LPos);
          TokType := TokSymbol;
          if(Token = '<')and(Line[LPos] = '=') then
          begin
             Token := '{';
             inc(LPos);
          end
          else if(Token = '>')and(Line[LPos] = '=') then
          begin
             Token := '}';
             inc(LPos);
          end
          else if(Token = '<')and(Line[LPos] = '>') then
          begin
             Token := '#';
             inc(LPos);
          end;
        end;
     '+','-','/','*','^':
       begin
        Token := Line[LPos];
        inc(LPos);
        while Line[LPos] in [' ',#9] do inc(LPos);
        if not(Line[LPos] in ['+','-']) then TokType := TokMath
        else Token := '';
       end;
     #1: TokType := TokEnd;
   end;
end;

procedure THIMathParseEx.Level0; //  
var i:integer;
begin
  LPos := 1; Err := -1;
  GetToken; Level1a; {$ifdef F_P} if Err>=0 then exit; {$endif}
  if TokType<>TokEnd then
    {$ifndef F_P}
    raise Exception.Create(e_Custom,'');
    {$else}Err:=0;{$endif};
end;

procedure THIMathParseEx.Level1a;  //   ( )
var
  op:char;
begin
  Level1b; {$ifdef F_P} if Err>=0 then exit; {$endif}
  while (Token = '<')or(Token = '>')or(Token = '{')or(Token = '}')or(Token = '=')or(Token = '#') do
   begin
    op := Token[1];
    GetToken; Level1b; {$ifdef F_P} if Err>=0 then exit; {$endif}
    case op of
     '<': FProgram.WriteByte(CmdSmall);
     '>': FProgram.WriteByte(CmdBig);
     '{': FProgram.WriteByte(CmdESmall);
     '}': FProgram.WriteByte(CmdEBig);
     '=': FProgram.WriteByte(CmdEqual);
     '#': FProgram.WriteByte(CmdNEqual);
    end;
   end;
end;

procedure THIMathParseEx.Level1b;  // 
var
  op:char;
  x2:real;
begin
  Level2; {$ifdef F_P} if Err>=0 then exit; {$endif}
  while (Token = '-')or(Token = '+') do
   begin
    op := Token[1];
    GetToken; Level2; {$ifdef F_P} if Err>=0 then exit; {$endif}
    if op = '+' then FProgram.WriteByte(CmdAdd)
    else FProgram.WriteByte(CmdSub);
   end;
end;

procedure THIMathParseEx.Level2;   // 
var op:char;
begin
  Level3; {$ifdef F_P} if Err>=0 then exit; {$endif}
  while (Token='/')or(Token='*')or(Token='div')or(Token='mod')do begin
    op :=  Token[1];
    GetToken; Level3; {$ifdef F_P} if Err>=0 then exit; {$endif}
    case op of
      '*': FProgram.WriteByte(CmdMul);
      '/': FProgram.WriteByte(CmdDiv);
      'd': FProgram.WriteByte(CmdIDiv);
      else FProgram.WriteByte(CmdIMod);
    end;
  end
end;

function IntPower(const Base:Extended; const Exponent:Integer):Extended;
asm
        fld1                      { Result := 1 }
        fld     Base
        test    eax,eax
        jg      @@2
        fdivr   ST,ST(1)          { Base := 1 / Base }
        neg     eax
        jnz     @@2
        jmp     @@3
@@1:    fmul    ST,ST             { X := Base * Base }
@@2:    shr     eax,1
        jnc     @@1
        fmul    ST(1),ST          { Result := Result * X }
        jnz     @@1
@@3:    fstp    st                { pop X from FPU stack }
        fwait
end;

procedure Power(var x, Exponent: real);
var y:integer;
begin
  y := Round(Exponent);
  if Exponent=y then x := IntPower(x, y)
  else if (x=0)and(Exponent>0) then x := 0
  else x := Exp(Exponent*Ln(x));
end;

procedure THIMathParseEx.Level3;   //   
begin
  Level4; {$ifdef F_P} if Err>=0 then exit; {$endif}
  while (Token = '^') do begin
    GetToken;
    Level4;
    {$ifdef F_P} if Err>=0 then exit; {$endif}
    FProgram.WriteByte(CmdPow);
  end;
end;

procedure THIMathParseEx.Level4;   //  
var op:char;
begin
  op := ' ';
  if (Token = '-')or(Token = '+') then begin
    op := Token[1];
    GetToken;
  end;
  Level5;
  if op = '-' then FProgram.WriteByte(CmdNeg);
end;

procedure THIMathParseEx.Level5;   //  
var op:char;
begin
  Level6; {$ifdef F_P} if Err>=0 then exit; {$endif}
  while (Token = 'and')or(Token = 'or')or(Token = 'xor')do begin
    op := Token[1];
    GetToken; Level6; {$ifdef F_P} if Err>=0 then exit; {$endif}
    case op of
      'a': FProgram.WriteByte(CmdAnd);
      'o': FProgram.WriteByte(CmdOr);
      'x': FProgram.WriteByte(CmdXor);
    end;
  end
end;

procedure THIMathParseEx.Level6;   // ,  
var i,j, cnt:integer; Y:real;
    Fd,FItem:TData; f:boolean;
    tmp:PData;
    Arr:PArray;
    Mtx:PMatrix;
begin
  if Token = '%' then begin
    //  ,  - syntax error
    GetToken;
    if TokType=TokNumber then begin
      i := Str2Int(Token)-1;
      if i=-1 then begin
        FProgram.WriteByte(CmdRead);
        FProgram.WriteByte(255);
      end else if i<FDataCount then begin
        GetToken;
        if Token = '[' then begin
        //     (  )
          GetToken; Level1a; {$ifdef F_P} if Err>=0 then exit; {$endif}
          if Token <> ',' then begin
          //   
            FProgram.WriteByte(CmdArr);
            FProgram.WriteByte(Byte(i));
            if Token <> ']' then
              {$ifdef F_P}
              begin Err:=0; exit; end;
              {$else}
              raise Exception.Create(e_Custom,'');
              {$endif}
          end
          else begin
          //   
            FProgram.WriteByte(CmdMatr);
            FProgram.WriteByte(Byte(i));
            GetToken; Level1a; {$ifdef F_P} if Err>=0 then exit; {$endif}
            if Token <> ']' then
              {$ifdef F_P}
              begin Err:=0; exit; end;
              {$else}
              raise Exception.Create(e_Custom,'');
              {$endif}
          end
        end
        else if Token = '(' then begin
        //   (  ???)
          FProgram.WriteByte(CmdUFunc);
          FProgram.WriteByte(Byte(i));
          GetToken;
          Level1a; {$ifdef F_P} if Err>=0 then exit; {$endif}
          cnt := 1;
          while Token = ',' do begin  //    
            GetToken;
            Level1a; {$ifdef F_P} if Err>=0 then exit; {$endif}
            inc(cnt);
          end;
          if Token <> ')' then
            {$ifdef F_P} begin Err:=0; exit; end;
            {$else}      raise Exception.Create(e_Custom,''); {$endif}
          FProgram.WriteInteger(cnt);
        end
        else begin
        //    .  
          FProgram.WriteByte(CmdRead);
          FProgram.WriteByte(Byte(i));
          exit;
        end
      end
      else{$ifdef F_P}begin Err:=0;exit;end{$else}raise Exception.Create(e_Custom,''){$endif};
    end
    else  {$ifdef F_P}begin Err:=0;exit;end{$else}raise Exception.Create(e_Custom,''){$endif};
  end
  else if Token = '(' then begin
    //  
    GetToken; Level1a; {$ifdef F_P} if Err>=0 then exit; {$endif}
    if Token <> ')' then
      {$ifdef F_P}
      begin Err:=0; exit; end;
      {$else}
      raise Exception.Create(e_Custom,'');
      {$endif}
  end
  else case TokType of
    TokName: begin if ReadFunc(LowerCase(Token)) then //   =>  
      else begin
        if (_prop_ExtNames>0)and(_prop_ExtNames<=FDataCount) then begin
          //     "  "
          FProgram.WriteByte(CmdUNFunc);
          FProgram.WriteString(Token);
          cnt := 0;
          GetToken;
          f := Token = '(';
          if f then begin
            inc(cnt);
            GetToken;
            Level1a; {$ifdef F_P} if Err>=0 then exit; {$endif}
            while Token = ',' do begin  //   
              GetToken;
              Level1a; {$ifdef F_P} if Err>=0 then exit; {$endif}
              inc(cnt);
            end;
            if Token <> ')' then
              {$ifdef F_P} begin Err:=0; exit; end;
              {$else}      raise Exception.Create(e_Custom,''); {$endif}
            FProgram.WriteInteger(cnt);
          end;
          if not f then exit; //  !
        end else
          {$ifdef F_P} begin Err:=0;if Err>=0 then exit; end;
          {$else}raise Exception.Create(e_Custom,''){$endif};
      end;
    end;
    //     
    TokReal,TokNumber: begin
      FProgram.WriteByte(CmdPush);
      FProgram.WriteReal(Str2Double(Token));
    end;
    
    TokHex: begin
      FProgram.WriteByte(CmdPush);
      FProgram.WriteReal(Hex2Int(Token));
    end;
    else  {$ifdef F_P}begin Err:=0; exit; end{$else}raise Exception.Create(e_Custom,''){$endif};
  end;
  GetToken;
end;

function Tan(const X: Extended): Extended;
//  Tan := Sin(X) / Cos(X)
asm
        FLD    X
        FPTAN
        FSTP   ST(0)      // FPTAN pushes 1.0 after result
        FWAIT
end;

function CoTan(const X: Extended): Extended;
// CoTan := Cos(X) / Sin(X) = 1 / Tan(X)
asm
        FLD   X
        FPTAN
        FDIVRP
        FWAIT
end;

function ArcTan2(const Y, X: Extended): Extended;
asm
        FLD     Y
        FLD     X
        FPATAN
        FWAIT
end;

function LogN(const Base, X: Extended): Extended;
// Log.N(X) := Log.2(X) / Log.2(N)
asm
        FLD1
        FLD     X
        FYL2X
        FLD1
        FLD     Base
        FYL2X
        FDIV
        FWAIT
end;

function THIMathParseEx.ReadFunc;
const FuncNames:array[1..34] of pchar = (
  'cos','sin','tg','ctg','arccos','arcsin','arctg','arcctg','atan',  //1..9
  'ch','sh','th','cth','arcch','arcsh','arcth','arccth',             //10..17
  'log','lg','ln','exp','sqr','sqrt','abs','sign',                   //18..25
  'round','frac','trunc','floor','ceil','min','max','odd','even');   //26..34
var i, cnt: integer;
begin
  Result := True;
  if      f = 'pi' then begin FProgram.WriteByte(CmdPush); FProgram.WriteReal(pi); exit; end
  else if f = 'e'  then begin FProgram.WriteByte(CmdPush); FProgram.WriteReal(digE); exit; end;
  Result := false;
  for I := 1 to length(FuncNames) do if FuncNames[I]=f then begin Result := True; break; end;
  if not Result then exit;
  
  GetToken;
  if Token <> '(' then
    {$ifdef F_P} begin Err:=0; exit; end;
    {$else}raise Exception.Create(e_Custom,'');{$endif}
  GetToken; Level1a; {$ifdef F_P} if Err>=0 then exit; {$endif}
  case I of
    9, 18:
        begin
         if Token<>',' then
           {$ifdef F_P}
           begin Err:=0; exit; end;
           {$else}
           raise Exception.Create(e_Custom,'');
           {$endif}
         GetToken; Level1a; {$ifdef F_P} if Err>=0 then exit; {$endif}
        end;
    
    26, 30:
        begin
          if Token=',' then begin
            GetToken; Level1a; {$ifdef F_P} if Err>=0 then exit; {$endif}
          end else begin
            FProgram.WriteByte(CmdPush);
            FProgram.WriteReal(1);
          end;
        end;
    31, 32:
        begin
          cnt := 0;
          while Token=',' do begin
            GetToken; Level1a; {$ifdef F_P} if Err>=0 then exit; {$endif}
            inc(cnt);
          end;
        end;
  end;
  if Token <> ')' then {$ifdef F_P}Err:=0{$else}raise Exception.Create(e_Custom,''){$endif};
  FProgram.WriteByte(CmdFunc);
  FProgram.WriteByte(Byte(i));
  if (i = 31) or (i = 32) then begin
    FProgram.WriteInteger(cnt);
  end;
end;

procedure THIMathParseEx.CalcFunc;
var x, y:real;
    j: integer;
    i: byte;
begin
  if not FProgram.ReadByte(i) then begin Err := 2; Exit; end;
  if not FStack.Pop(x) then begin Err := 1; Exit; end;
  case I of
    1:  x := cos(x*AngleMode);
    2:  x := sin(x*AngleMode);
    3:  {$ifdef F_P}if cos(x*AngleMode)=0 then begin Err:=1;exit;end else{$endif}x := Tan(x*AngleMode);
    4:  {$ifdef F_P}if sin(x*AngleMode)=0 then begin Err:=1;exit;end else{$endif}x := CoTan(x*AngleMode);
    5:  {$ifdef F_P}if (x>1)or(x<-1) then begin Err:=1;exit;end else{$endif}x := ArcTan2(sqrt(1-x*x),x)/AngleMode;
    6:  {$ifdef F_P}if (x>1)or(x<-1) then begin Err:=1;exit;end else{$endif}x := ArcTan2(x,sqrt(1-x*x))/AngleMode;
    7:  x := ArcTan2(x,1)/AngleMode;
    8:  x := ArcTan2(1,x)/AngleMode;
    9:  begin
         y := x;
         if not FStack.Pop(x) then begin Err := 1; Exit; end;
         x := ArcTan2(x,y)/AngleMode;
        end;
    10: begin y := exp(x); x := (y+1/y)/2 end;
    11: begin y := exp(x); x := (y-1/y)/2 end;
    12: begin y := exp(2*x); x := (y-1)/(y+1) end;
    13: {$ifdef F_P}if x=0 then begin Err:=1;exit;end else{$endif}begin y := exp(2*x); x := (y+1)/(y-1) end;
    14: {$ifdef F_P}if x<1 then begin Err:=1;exit;end else{$endif}x := LogN(digE,x+sqrt(x*x-1));
    15: x := LogN(digE,x+sqrt(x*x+1));
    16: {$ifdef F_P}if (x>=1)or(x<=-1) then begin Err:=1;exit;end else{$endif}x := LogN(digE,(1+x)/(1-x))/2;
    17: {$ifdef F_P}if (x<=1)and(x>=-1)then begin Err:=1;exit;end else{$endif}x := LogN(digE,(x+1)/(x-1))/2;
    18: begin
          y := x;
          if not FStack.Pop(x) then begin Err := 1; Exit; end;
          {$ifdef F_P}if (x<=0)or(y<=0) then begin Err:=1;exit;end;{$endif}
          x := LogN(x,y)
        end;
    19: {$ifdef F_P}if x<=0 then begin Err:=1;exit;end else{$endif}x := LogN(10,x);
    20: {$ifdef F_P}if x<=0 then begin Err:=1;exit;end else{$endif}x := LogN(digE,x);
    21: x := exp(x);
    22: x := sqr(x);
    23: {$ifdef F_P}if x<0 then begin Err:=1;exit;end else{$endif}x := sqrt(x);
    24: if x<0 then x := -x;
    25: if x<0 then x := -1 else if x>0 then x:=1;
    26..30:
        begin
          y := x;
          if not FStack.Pop(x) then begin Err := 1; Exit; end;
          x := x/y;
          case I of
            26:  x := y*round(x);
            27:  x := y*frac (x);
            28:  x := y*trunc(x);
            29:  if frac(x)<0 then x := y*(trunc(x)-1) //floor
                 else x := y*trunc(x);
            else if frac(x)>0 then x := y*(trunc(x)+1) //ceil
                 else x := y*trunc(x);
          end;
        end;
    31,32:
      begin
        if not FProgram.ReadInteger(j) then begin Err := 2; Exit; end;
        while j > 0 do begin
          if not FStack.Pop(y) then begin Err := 1; Exit; end;
          case I of
            31:  if x>y then x := y;
            else if x<y then x := y;
          end;
          dec(i);
        end;
      end;
    33: x := ord(odd(round(x)));     {odd}
    34: x := ord(not odd(round(x))); {even}                   
  end;
  FStack.Push(x);
end;

end.
